# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

import numpy
from Crypto.Cipher import AES

from .symmetric import SymmetricKey
from .utils import base64_encode, base64_decode


class AESKey(SymmetricKey):
    """ AES Key """

    def __new__(cls, key: dict):
        """
        Create AES key

        :param key: key info
        :return: AESKey object
        """
        if key is None:
            return None
        elif cls is AESKey:
            if isinstance(key, AESKey):
                # return AESKey object directly
                return key
        # new AESKey(dict)
        return super().__new__(cls, key)

    def __init__(self, key: dict):
        if self is key:
            # no need to init again
            return
        super().__init__(key)
        # key data
        data = self.get('data')
        if data is None:
            # generate data and iv
            data = bytes(numpy.random.bytes(self.size))
            self['data'] = base64_encode(data)
            iv = bytes(numpy.random.bytes(AES.block_size))
            self['iv'] = base64_encode(iv)
            # self['mode'] = 'CBC'
            # self['padding'] = 'PKCS7'
        else:
            data = base64_decode(data)
        # initialization vector
        iv = self.get('iv')
        if iv is None:
            # iv = b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0'
            iv = AES.block_size * chr(0).encode('utf-8')
            self['iv'] = base64_encode(iv)
        else:
            iv = base64_decode(iv)
        self.data = data
        self.iv = iv

    @property
    def size(self) -> int:
        size = self.get('size')
        if size is None:
            return 32
        else:
            return int(size)

    @staticmethod
    def __pad(data: bytes) -> bytes:
        block_size = AES.block_size
        amount = block_size - len(data) % block_size
        if amount == 0:
            amount = block_size
        pad = chr(amount).encode('utf-8')
        return data + pad * amount

    @staticmethod
    def __unpad(data: bytes) -> bytes:
        amount = data[-1]
        return data[:-amount]

    def encrypt(self, data: bytes) -> bytes:
        data = self.__pad(data)
        key = AES.new(self.data, AES.MODE_CBC, self.iv)
        return key.encrypt(data)

    def decrypt(self, data: bytes) -> Optional[bytes]:
        key = AES.new(self.data, AES.MODE_CBC, self.iv)
        plaintext = key.decrypt(data)
        return self.__unpad(plaintext)


# register symmetric key class with algorithm
SymmetricKey.register(algorithm=SymmetricKey.AES, key_class=AESKey)        # default
SymmetricKey.register(algorithm='AES/CBC/PKCS7Padding', key_class=AESKey)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABCMeta
from enum import IntEnum

from .crypto.utils import sha256, ripemd160, base58_encode, base58_decode


class NetworkID(IntEnum):
    """
        @enum MKMNetworkID

        @abstract A network type to indicate what kind the entity is.

        @discussion An address can identify a person, a group of people,
            a team, even a thing.

            MKMNetwork_Main indicates this entity is a person's account.
            An account should have a public key, which proved by meta data.

            MKMNetwork_Group indicates this entity is a group of people,
            which should have a founder (also the owner), and some members.

            MKMNetwork_Moments indicates a special personal social network,
            where the owner can share information and interact with its friends.
            The owner is the king here, it can do anything and no one can stop it.

            MKMNetwork_Polylogue indicates a virtual (temporary) social network.
            It's created to talk with multi-people (but not too much, e.g. < 100).
            Any member can invite people in, but only the founder can expel member.

            MKMNetwork_Chatroom indicates a massive (persistent) social network.
            It's usually more than 100 people in it, so we need administrators
            to help the owner to manage the group.

            MKMNetwork_SocialEntity indicates this entity is a social entity.

            MKMNetwork_Organization indicates an independent organization.

            MKMNetwork_Company indicates this entity is a company.

            MKMNetwork_School indicates this entity is a school.

            MKMNetwork_Government indicates this entity is a government department.

            MKMNetwork_Department indicates this entity is a department.

            MKMNetwork_Thing this is reserved for IoT (Internet of Things).

        Bits:
            0000 0001 - this entity's branch is independent (clear division).
            0000 0010 - this entity can contains other group (big organization).
            0000 0100 - this entity is top organization.
            0000 1000 - (Main) this entity acts like a human.

            0001 0000 - this entity contains members (Group)
            0010 0000 - this entity needs other administrators (big organization)
            0100 0000 - this is an entity in reality.
            1000 0000 - (IoT) this entity is a 'Thing'.

            (All above are just some advices to help choosing numbers :P)
    """
    ################################
    #  BTC Network
    ################################
    BTCMain = 0x00      # 0000 0000 (BitCoin Address)
    # BTCTest = 0x6f    # 0110 1111 (BitCoin Test)

    ################################
    #  Person Account
    ################################
    Main = 0x08         # 0000 1000 (Person)

    ################################
    #  Virtual Groups
    ################################
    Group = 0x10        # 0001 0000 (Multi-Persons)
    # Moments = 0x18    # 0001 1000 (Twitter)
    Polylogue = 0x10    # 0001 0000 (Multi-Persons Chat, N < 100)
    Chatroom = 0x30     # 0011 0000 (Multi-Persons Chat, N >= 100)

    ################################
    #  Social Entities in Reality
    ################################
    # SocialEntity = 0x50  # 0101 0000
    # Organization = 0x74  # 0111 0100
    # Company = 0x76       # 0111 0110
    # School = 0x77        # 0111 0111
    # Government = 0x73    # 0111 0011
    # Department = 0x52    # 0101 0010

    ################################
    #  Network
    ################################
    Provider = 0x76        # 0111 0110 (Service Provider)
    Station = 0x88         # 1000 1000 (Server Node)

    ################################
    #  Internet of Things
    ################################
    Thing = 0x80           # 1000 0000 (IoT)
    Robot = 0xC8           # 1100 1000

    def is_user(self) -> bool:
        return (self.value & self.Main.value) or (self.value == self.BTCMain.value)

    def is_person(self) -> bool:
        return (self.value == self.Main.value) or (self.value == self.BTCMain.value)

    def is_group(self) -> bool:
        return self.value & self.Group.value

    def is_station(self) -> bool:
        return self.value == self.Station.value

    def is_provider(self) -> bool:
        return self.value == self.Provider.value

    def is_thing(self) -> bool:
        return self.value & self.Thing.value

    def is_robot(self) -> bool:
        return self.value == self.Robot.value


class Address(str, metaclass=ABCMeta):
    """This class is used to build address for ID

        Address for MKM ID
        ~~~~~~~~~~~~~~~~~~

        properties:
            network - address type
            number  - search number
    """

    def __new__(cls, address: str):
        """
        Create address object with string

        :param address: address string
        :return: Address object
        """
        if address is None:
            return None
        elif cls is Address:
            if isinstance(address, Address):
                # return Address object directly
                return address
            # Address for broadcast
            length = len(address)
            # anywhere
            if length == len(ANYWHERE) and address.lower() == ANYWHERE:
                return ANYWHERE
            # everywhere
            if length == len(EVERYWHERE) and address.lower() == EVERYWHERE:
                return EVERYWHERE
            # try to create address object
            for clazz in cls.address_classes():
                try:
                    # noinspection PyTypeChecker
                    return clazz.__new__(clazz, address)
                except ValueError:
                    continue
            raise ValueError('unrecognized address: %s' % address)
        # subclass
        return super().__new__(cls, address)

    @property
    def network(self) -> NetworkID:
        """
        Get network type of address

        :return: NetworkID
        """
        yield None

    @property
    def number(self) -> int:
        """
        Get search number of address

        :return: search number [0, 2^32)
        """
        return 0

    @property
    def is_broadcast(self) -> bool:
        assert self.network is not None, 'address error: %s' % self
        value = self.network.value
        if value == NetworkID.Group.value:
            # group address
            return self == EVERYWHERE
        elif value == NetworkID.Main.value:
            # user address
            return self == ANYWHERE

    #
    #   Runtime
    #
    __address_classes = []  # class list

    @classmethod
    def register(cls, address_class) -> bool:
        """
        Register address class

        :param address_class: class for parsing ID.address
        :return: False on error
        """
        cls.__address_classes.append(address_class)
        # TODO: check issubclass(address_class, Address)
        return True

    @classmethod
    def address_classes(cls) -> list:
        """
        Get content class with type

        :return: all address classes
        """
        return cls.__address_classes


"""
    Address like BitCoin
    ~~~~~~~~~~~~~~~~~~~~

    data format: "network+digest+check_code"
        network    --  1 byte
        digest     -- 20 bytes
        check_code --  4 bytes

    algorithm:
        fingerprint = sign(seed, SK);
        digest      = ripemd160(sha256(fingerprint));
        check_code  = sha256(sha256(network + digest)).prefix(4);
        address     = base58_encode(network + digest + check_code);
"""


def check_code(data: bytes) -> bytes:
    # check code in BTC address
    return sha256(sha256(data))[:4]


def user_number(code: bytes) -> int:
    # get user number, which for remembering and searching user
    return int.from_bytes(code, byteorder='little')


class DefaultAddress(Address):

    def __new__(cls, address: str):
        if address is None:
            return None
        elif cls is DefaultAddress:
            if isinstance(address, DefaultAddress):
                # return DefaultAddress directly
                return address
        # new DefaultAddress(str)
        return super().__new__(cls, address)

    def __init__(self, address: str):
        if self is address:
            # no need to init again
            return
        super().__init__()
        # get fields from string
        prefix_digest_code = base58_decode(address)
        if len(prefix_digest_code) != 25:
            raise ValueError('BTC address length error')
        # split them
        prefix = prefix_digest_code[:1]
        digest = prefix_digest_code[1:-4]
        code = prefix_digest_code[-4:]
        # check them
        if check_code(prefix + digest) != code:
            raise ValueError('BTC address check code error')
        network = ord(prefix)
        self.__network = NetworkID(network)
        self.__number = user_number(code)

    @property
    def network(self) -> NetworkID:
        return self.__network

    @property
    def number(self) -> int:
        return self.__number

    #
    #   Factory
    #
    @classmethod
    def new(cls, data: bytes, network: NetworkID=0) -> Address:
        """
        Generate address with fingerprint and network ID

        :param data:    fingerprint (signature/key.data)
        :param network: address type
        :return:        Address object
        """
        prefix = chr(network).encode('latin1')
        digest = ripemd160(sha256(data))
        code = check_code(prefix + digest)
        address = base58_encode(prefix + digest + code)
        return cls(address)


# register default address class
Address.register(DefaultAddress)  # default


"""
    Address for broadcast
    ~~~~~~~~~~~~~~~~~~~~~
"""


class ConstantAddress(Address):

    def __new__(cls, address: str, network: NetworkID, number: int):
        self = super().__new__(cls, address)
        self.__network = network
        self.__number = number
        return self

    @property
    def network(self) -> NetworkID:
        return self.__network

    @property
    def number(self) -> int:
        return self.__number


ANYWHERE = ConstantAddress(address="anywhere", network=NetworkID.Main, number=9527)
EVERYWHERE = ConstantAddress(address="everywhere", network=NetworkID.Group, number=9527)



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Utilities
    ~~~~~~~~~

    Crypto utilities
"""

import hashlib
from typing import Optional

import base58
import base64


def sha256(data: bytes) -> bytes:
    """ SHA-256 """
    # return SHA256.new(data).digest()
    return hashlib.sha256(data).digest()


def ripemd160(data: bytes) -> bytes:
    """ RIPEMD-160 """
    hash_obj = hashlib.new('ripemd160')
    hash_obj.update(data)
    return hash_obj.digest()


def base58_encode(data: bytes) -> str:
    """ BASE-58 Encode """
    return base58.b58encode(data).decode('utf-8')


def base58_decode(string: str) -> Optional[bytes]:
    """ BASE-58 Decode """
    return base58.b58decode(string)


def base64_encode(data: bytes) -> str:
    """ BASE-64 Encode """
    return base64.b64encode(data).decode('utf-8')


def base64_decode(string: str) -> Optional[bytes]:
    """ BASE-64 Decode """
    return base64.b64decode(string)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from .address import Address, NetworkID, ANYWHERE, EVERYWHERE


class ID(str):
    """
        ID for entity (User/Group)
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        data format: "name@address[/terminal]"

        fields:
            name     - entity name, the seed of fingerprint to generate address
            address  - a string to identify an entity
            terminal - entity login resource(device), OPTIONAL
    """

    def __new__(cls, identifier: str):
        """
        Create ID object with string

        :param identifier: ID string with format 'name@address/terminal'
        :return: ID object
        """
        if identifier is None:
            return None
        elif cls is ID:
            if isinstance(identifier, ID):
                # return ID object directly
                return identifier
        # new ID(str)
        return super().__new__(cls, identifier)

    def __init__(self, identifier: str):
        if self is identifier:
            # no need to init again
            return
        super().__init__()
        # lazy
        self.__name: str = None
        self.__address: Address = None
        self.__terminal: str = None

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        if super().__eq__(other):
            return True
        identifier = ID(other)
        assert identifier.valid, 'other ID not valid: %s' % other
        return self.name == identifier.name and self.address == identifier.address

    def __hash__(self) -> int:
        string = str(self.__address)
        if self.__name is not None and len(self.__name) > 0:
            string = self.__name + '@' + string
        if self.__terminal is not None and len(self.__terminal) > 0:
            string = string + '/' + self.__terminal
        return hash(string)

    @property
    def name(self) -> Optional[str]:
        if self.valid:
            return self.__name

    @property
    def address(self) -> Address:
        if self.__address is None:
            # split ID string
            pair = self.split('/', 1)
            if len(pair) == 2:
                # got terminal
                self.__terminal = pair[1]
            else:
                self.__terminal = ''
            pair = pair[0].split('@', 1)
            if len(pair) == 2:
                # got name & address
                self.__name = pair[0]
                self.__address = Address(pair[1])
            else:
                # got address
                self.__name = ''
                self.__address = Address(pair[0])
        return self.__address

    @property
    def terminal(self) -> Optional[str]:
        if self.valid:
            return self.__terminal

    @property
    def type(self) -> NetworkID:
        """ ID type """
        address = self.address
        if address is not None:
            return address.network

    @property
    def number(self) -> int:
        """ Search number of this ID """
        address = self.address
        if address is not None:
            return address.number

    @property
    def valid(self) -> bool:
        address = self.address
        if address is not None:
            return address.number > 0

    @property
    def is_broadcast(self) -> bool:
        assert self.address is not None, 'ID error: %s' % self
        return self.address.is_broadcast

    #
    #   Factory
    #
    @classmethod
    def new(cls, address: Address, name: str=None, terminal: str=None):
        # concatenate ID string
        string = str(address)
        if name is not None:
            string = name + '@' + string
        if terminal is not None:
            string = string + '/' + terminal
        # new ID(str)
        identifier = cls(string)
        identifier.__name = name
        identifier.__address = address
        identifier.__terminal = terminal
        return identifier


"""
    ID for broadcast
    ~~~~~~~~~~~~~~~~
"""


ANYONE = ID.new(name="anyone", address=ANYWHERE)
EVERYONE = ID.new(name="everyone", address=EVERYWHERE)



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5 as Cipher_PKCS1_v1_5
from Crypto.Signature import PKCS1_v1_5 as Signature_PKCS1_v1_5

from .asymmetric import PublicKey, PrivateKey


class RSAPublicKey(PublicKey):
    """ RSA Public Key """

    def __new__(cls, key: dict):
        """
        Create RSA public key

        :param key: key info
        :return: RSAPublicKey object
        """
        if key is None:
            return None
        elif cls is RSAPublicKey:
            if isinstance(key, RSAPublicKey):
                # return RSAPublicKey object directly
                return key
        # new RSAPublicKey(dict)
        return super().__new__(cls, key)

    def __init__(self, key: dict):
        if self is key:
            # no need to init again
            return
        super().__init__(key)
        # data in 'PEM' format
        data = key['data']
        self.key = RSA.importKey(data)
        self.data = self.key.exportKey(format='DER')

    # noinspection PyTypeChecker
    def encrypt(self, data: bytes) -> bytes:
        cipher = Cipher_PKCS1_v1_5.new(self.key)
        return cipher.encrypt(data)

    def verify(self, data: bytes, signature: bytes) -> bool:
        hash_obj = SHA256.new(data)
        verifier = Signature_PKCS1_v1_5.new(self.key)
        try:
            return verifier.verify(hash_obj, signature)
        except ValueError:
            # raise ValueError("Invalid signature")
            return False


class RSAPrivateKey(PrivateKey):
    """ RSA Private Key """

    def __new__(cls, key: dict):
        """
        Create RSA private key

        :param key: key info
        :return: RSAPrivateKey object
        """
        if key is None:
            return None
        elif cls is RSAPrivateKey:
            if isinstance(key, RSAPrivateKey):
                # return RSAPrivateKey object directly
                return key
        # new RSAPrivateKey(dict)
        return super().__new__(cls, key)

    def __init__(self, key: dict):
        if self is key:
            # no need to init again
            return
        super().__init__(key)
        # data in 'PEM' format
        data: str = key.get('data')
        if data is None:
            # generate private key data
            private_key = RSA.generate(self.size)
            data: bytes = private_key.exportKey()
            self['data'] = data.decode('utf-8')
            self['mode'] = 'ECB'
            self['padding'] = 'PKCS1'
            self['digest'] = 'SHA256'
        else:
            tag1 = '-----BEGIN RSA PRIVATE KEY-----'
            tag2 = '-----END RSA PRIVATE KEY-----'
            pos2 = data.find(tag2)
            if pos2 > 0:
                pos1 = data.find(tag1)
                data = data[pos1: pos2 + len(tag2)]
        # create key
        self.key = RSA.importKey(data)

    @property
    def size(self):
        bits = self.get('keySizeInBits')
        if bits is None:
            return 1024
        else:
            return int(bits)

    @property
    def public_key(self) -> PublicKey:
        pk = self.key.publickey()
        data = pk.exportKey()
        info = {
            'algorithm': PublicKey.RSA,
            'data': data.decode('utf-8'),
            'mode': 'ECB',
            'padding': 'PKCS1',
            'digest': 'SHA256'
        }
        return RSAPublicKey(info)

    # noinspection PyTypeChecker
    def decrypt(self, data: bytes) -> Optional[bytes]:
        cipher = Cipher_PKCS1_v1_5.new(self.key)
        sentinel = ''
        plaintext = cipher.decrypt(data, sentinel)
        if sentinel:
            print('error: ' + sentinel)
        return plaintext

    def sign(self, data: bytes) -> bytes:
        hash_obj = SHA256.new(data)
        signer = Signature_PKCS1_v1_5.new(self.key)
        return signer.sign(hash_obj)


# register public key class with algorithm
PublicKey.register(algorithm=PublicKey.RSA, key_class=RSAPublicKey)             # default
PublicKey.register(algorithm='SHA256withRSA', key_class=RSAPublicKey)
PublicKey.register(algorithm='RSA/ECB/PKCS1Padding', key_class=RSAPublicKey)

# register private key class with algorithm
PrivateKey.register(algorithm=PrivateKey.RSA, key_class=RSAPrivateKey)          # default
PrivateKey.register(algorithm='SHA256withRSA', key_class=RSAPrivateKey)
PrivateKey.register(algorithm='RSA/ECB/PKCS1Padding', key_class=RSAPrivateKey)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Disk Operating System
    ~~~~~~~~~~~~~~~~~~~~~

    File access
"""

import json
import os
from typing import Union, Optional


class File:

    def __init__(self, path: str):
        super().__init__()
        self.__path = path
        self.__data: Union[bytes, str] = None

    @classmethod
    def exists(cls, path: str) -> bool:
        return os.path.exists(path)

    @classmethod
    def make_dirs(cls, directory: str) -> bool:
        if cls.exists(directory):
            # directory exists
            return os.path.isdir(directory)
        try:
            os.makedirs(directory)
            return True
        except IOError:
            return False

    def read(self, mode: str='rb') -> Union[bytes, str, None]:
        if self.__data is not None:
            # get data from cache
            return self.__data
        if not self.exists(self.__path):
            # file not found
            return None
        if not os.path.isfile(self.__path):
            # the path is not a file
            raise IOError('%s is not a file' % self.__path)
        with open(self.__path, mode) as file:
            self.__data = file.read()
        return self.__data

    def write(self, data: Union[bytes, str], mode: str='wb') -> bool:
        directory = os.path.dirname(self.__path)
        if not self.make_dirs(directory):
            return False
        with open(self.__path, mode) as file:
            if len(data) == file.write(data):
                # OK, update cache
                self.__data = data
                return True

    def append(self, data: Union[bytes, str], mode: str='ab') -> bool:
        if not self.exists(self.__path):
            # new file
            return self.write(data=data, mode=mode)
        # append to exists file
        with open(self.__path, mode) as file:
            if len(data) == file.write(data):
                # OK, erase cache
                self.__data = None
                return True

    def remove(self, path: str=None) -> bool:
        if path is None:
            path = self.__path
        if self.exists(path):
            os.remove(path)
            return True


class TextFile(File):

    def read(self, mode: str='r') -> Optional[str]:
        return super().read(mode=mode)

    def write(self, text: str, mode: str='w') -> bool:
        return super().write(data=text, mode=mode)

    def append(self, text: str, mode: str='a') -> bool:
        return super().append(data=text, mode=mode)


class JSONFile(TextFile):

    def __init__(self, path: str):
        super().__init__(path=path)
        self.__container: Union[dict, list] = None

    def read(self, **kwargs) -> Union[dict, list, None]:
        if self.__container is not None:
            # get content from cache
            return self.__container
        # read as text file
        text = super().read()
        if text is not None:
            # convert text string to JsON object
            self.__container = json.loads(text)
        return self.__container

    def write(self, container: Union[dict, list], **kwargs) -> bool:
        # convert JsON object to text string
        text = json.dumps(container)
        if text is None:
            raise ValueError('cannot convert to JsON: %s' % container)
        if super().write(text=text):
            # OK, update cache
            self.__container = container
            return True

    def append(self, text: str, mode: str='a') -> bool:
        raise AssertionError('JsON file cannot append')



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Immortals
    ~~~~~~~~~

    Built-in users
"""

import os
from typing import Optional

from .crypto import PrivateKey
from .address import NetworkID
from .identifier import ID
from .meta import Meta
from .profile import Profile
from .user import User, LocalUser, UserDataSource
from .dos import JSONFile


def load_resource_file(filename: str) -> dict:
    directory = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(directory, 'res', filename)
    file = JSONFile(path=path)
    return file.read()


class Immortals(UserDataSource):

    def __init__(self):
        super().__init__()
        # caches
        self.__ids = {}
        self.__metas = {}
        self.__private_keys = {}
        self.__profiles = {}
        self.__users = {}
        # load built-in users
        self.__load_user('hulk')
        self.__load_user('moki')

    def __load_user(self, username: str):
        # load meta
        meta = Meta(load_resource_file(filename=username+'_meta.js'))
        if meta is None:
            # meta not found
            return False
        # generate ID
        identifier = meta.generate_identifier(network=NetworkID.Main)
        # cache them
        self.cache_identifier(identifier=identifier)
        self.cache_meta(meta=meta, identifier=identifier)
        # load private key
        private_key = PrivateKey(load_resource_file(filename=username+'_secret.js'))
        if private_key is not None:
            # cache key
            self.cache_private_key(private_key=private_key, identifier=identifier)
        # load profile
        profile = Profile(load_resource_file(filename=username+'_profile.js'))
        if profile is None:
            # profile not found
            return True
        if self.verify_profile(profile=profile):
            # profile contains signature
            return True
        if private_key is None:
            return False
        # copy 'name'
        names = profile.get('names')
        if names is None:
            name = profile.get('name')
            if name is not None:
                profile.set_property('name', name)
        elif len(names) > 0:
            name = names[0]
            profile.set_property('name', name)
        # copy 'avatar'
        avatar = profile.get('avatar')
        if avatar is not None:
            profile.set_property('avatar', avatar)
        # sign and cache
        self.sign_profile(profile=profile)
        self.cache_profile(profile=profile)
        return True

    def cache_identifier(self, identifier: ID) -> bool:
        self.__ids[identifier] = identifier
        return True

    def cache_meta(self, meta: Meta, identifier: ID) -> bool:
        if not meta.match_identifier(identifier):
            return False
        self.__metas[identifier] = meta
        return True

    def cache_private_key(self, private_key: PrivateKey, identifier: ID) -> bool:
        self.__private_keys[identifier] = private_key
        return True

    def cache_profile(self, profile: Profile) -> bool:
        if not self.verify_profile(profile=profile):
            return False
        self.__profiles[profile.identifier] = profile
        return True

    def verify_profile(self, profile: Profile) -> bool:
        if profile.valid:
            return True
        identifier = self.identifier(profile.identifier)
        meta = self.meta(identifier=identifier)
        if meta is None:
            return False
        return profile.verify(public_key=meta.key)

    def sign_profile(self, profile: Profile) -> Optional[Profile]:
        if profile.valid:
            return profile
        identifier = self.identifier(profile.identifier)
        key = self.private_key_for_signature(identifier)
        if key is None:
            return None
        profile.sign(private_key=key)
        return profile

    def identifier(self, string: str) -> Optional[ID]:
        return self.__ids.get(string)

    def user(self, identifier: ID) -> Optional[User]:
        #  get from barrack cache
        user = self.__users.get(identifier)
        if user is not None:
            return user
        # check meta and private key
        meta = self.meta(identifier=identifier)
        if meta is not None:
            key = self.private_key_for_signature(identifier=identifier)
            if key is None:
                user = User(identifier=identifier)
            else:
                user = LocalUser(identifier=identifier)
            # cache it in barrack
            user.delegate = self
            self.__users[identifier] = user
            return user

    #
    #   EntityDataSource
    #
    def meta(self, identifier: ID) -> Optional[Meta]:
        return self.__metas.get(identifier)

    def profile(self, identifier: ID) -> Optional[Profile]:
        return self.__profiles.get(identifier)

    #
    #   UserDataSource
    #
    def private_key_for_signature(self, identifier: ID) -> Optional[PrivateKey]:
        return self.__private_keys.get(identifier)

    def private_keys_for_decryption(self, identifier: ID) -> Optional[list]:
        key = self.__private_keys.get(identifier)
        if key is not None:
            return [key]

    def contacts(self, identifier: ID) -> Optional[list]:
        if identifier not in self.__ids:
            return None
        array = []
        for key, value in self.__ids.items():
            if key == identifier:
                continue
            array.append(value)
        return array



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABCMeta, abstractmethod
from typing import Optional

from .cryptography import CryptographyKey


class SymmetricKey(CryptographyKey, metaclass=ABCMeta):
    """This class is used to encrypt or decrypt message data

        Symmetric Cryptography Key
        ~~~~~~~~~~~~~~~~~~~~~~~~~~

        key data format: {
            algorithm : "AES", // "DES", ...
            data      : "{BASE64_ENCODE}",
            ...
        }
    """

    AES = 'AES'
    DES = 'DES'

    # noinspection PyTypeChecker
    def __new__(cls, key: dict):
        """
        Create symmetric key

        :param key: key info (with algorithm='AES')
        :return: symmetric key
        """
        if key is None:
            return None
        elif cls is SymmetricKey:
            if isinstance(key, SymmetricKey):
                # return SymmetricKey object directly
                return key
            # get class by algorithm name
            clazz = cls.key_class(algorithm=key['algorithm'])
            if clazz is not None:
                assert issubclass(clazz, SymmetricKey), '%s must be sub-class of SymmetricKey' % clazz
                return clazz.__new__(clazz, key)
            else:
                raise ModuleNotFoundError('Invalid key algorithm: %s' % key)
        # subclass
        return super().__new__(cls, key)

    def __eq__(self, other) -> bool:
        if not isinstance(other, SymmetricKey):
            return False
        if super().__eq__(other):
            return True
        promise = 'Moky loves May Lee forever!'.encode('utf-8')
        return self.decrypt(other.encrypt(promise)) == promise

    @abstractmethod
    def encrypt(self, data: bytes) -> bytes:
        """
        ciphertext = encrypt(plaintext, PW)

        :param data: plaintext
        :return:     ciphertext
        """
        pass

    @abstractmethod
    def decrypt(self, data: bytes) -> Optional[bytes]:
        """
        plaintext = decrypt(ciphertext, PW);

        :param data: ciphertext
        :return:     plaintext
        """
        pass

    #
    #   Runtime
    #
    __key_classes = {}  # class map

    @classmethod
    def register(cls, algorithm: str, key_class=None) -> bool:
        """
        Register symmetric key class with algorithm

        :param algorithm: key algorithm
        :param key_class: if key class is None, then remove with algorithm
        :return: False on error
        """
        if key_class is None:
            cls.__key_classes.pop(algorithm, None)
        else:
            cls.__key_classes[algorithm] = key_class
        # TODO: check issubclass(key_class, SymmetricKey)
        return True

    @classmethod
    def key_class(cls, algorithm: str):
        """
        Get symmetric key class with algorithm

        :param algorithm: key algorithm
        :return: symmetric key class
        """
        return cls.__key_classes.get(algorithm)
